"""OAuth 2.0 WSGI server middleware to request a MyProxy certificates using a
resource server request
"""
__author__ = "R B Wilkinson"
__date__ = "19/03/12"
__copyright__ = "(C) 2012 Science and Technology Facilities Council"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id$"

import httplib
import logging

from webob import Request

import ndg.oauth.client.lib.certificate_request as certificate_request
from ndg.oauth.client.lib.oauth2client import Oauth2Client
from ndg.httpsclient.ssl_context_util import SSlContextConfig

log = logging.getLogger(__name__)

class CertificateRequestMiddleware(object):
    """
    OAuth 2.0 client middleware that requests a certificate using a resource
    server request
    """
    PARAM_PREFIX = 'certreq.'
    CERTIFICATE_ENVIRON_KEY_OPTION = 'oauth2_certificate_environ_key'
    CERTIFICATE_SESSION_KEY_OPTION = 'oauth2_certificate_session_key'
    CERTIFICATE_REQUEST_PARAMETER_OPTION = 'certificate_request_parameter'
    RESOURCE_SERVER_URL_OPTION = 'resource_server_url'
    SESSION_KEY_OPTION = 'session_key'
    TOKEN_KEY_OPTION = 'oauth2_token_key'
    PROPERTY_DEFAULTS = {
        CERTIFICATE_ENVIRON_KEY_OPTION: 'oauth2client.certificate',
        CERTIFICATE_SESSION_KEY_OPTION: 'oauth2client.certificate',
        CERTIFICATE_REQUEST_PARAMETER_OPTION: 'certificate_request',
        RESOURCE_SERVER_URL_OPTION: '',
        SESSION_KEY_OPTION: 'beaker.session.oauth2client',
        TOKEN_KEY_OPTION: 'oauth2client.token',
    }

    def __init__(self, app, app_conf, prefix=PARAM_PREFIX, **local_conf):
        """
        @param app: wrapped application/middleware
        @type app: WSGI application
        @param app_conf: application configuration settings - ignored - this
        method includes this arg to fit Paste middleware / app function 
        signature
        @type app_conf: dict
        @param prefix: optional prefix for parameter names included in the 
        local_conf dict - enables these parameters to be filtered from others
        which don't apply to this middleware
        @param local_conf: attribute settings to apply
        @type local_conf: dict
        """
        self._app = app
        self._set_configuration(prefix, local_conf)

    def __call__(self, environ, start_response):
        """
        @param environ: WSGI environment
        @type environ: dict
        @param start_response: WSGI start response function
        @type start_response: 
        @return: WSGI response
        @rtype: iterable
        """
        log.debug("CertificateRequestMiddleware.__call__ ...")

        req = Request(environ)
        log.debug("Request url: %s", req.url)
        log.debug("Request host_url: %s", req.host_url)
        log.debug("Request application_url: %s", req.application_url)

        # Get session.
        session = environ.get(self.session_env_key)
        if session is None:
            raise Exception(
                'CertificateRequestMiddleware.__call__: No beaker session key '
                '"%s" found in environ' % self.session_env_key)

        # Determine whether a certificate is stored in the session already.
        key_cert = None
        if self.__class__.CERTIFICATE_SESSION_KEY_OPTION in session:
            key_cert = session[self.__class__.CERTIFICATE_SESSION_KEY_OPTION]
        else:
            # If not, and an access token has been obtained, request a
            # certificate.
            client = Oauth2Client.get_client_instance(session, None,
                                                      create=False)
            if client and client.access_token:
                key_cert = certificate_request.request_certificate(
                                            client.access_token,
                                            self.resource_server_url,
                                            self.ssl_config,
                                            self.certificate_request_parameter)
                session[self.__class__.CERTIFICATE_SESSION_KEY_OPTION
                                                                    ] = key_cert
                # Make certificate available in the session immediately.
                session.save()
                session.persist()

        if key_cert:
            environ[self.certificate_env_key] = key_cert

        app_iter = self._app(environ, start_response)
        return app_iter

    def _set_configuration(self, prefix, local_conf):
        """Sets the configuration values.

        @param prefix: optional prefix for parameter names included in the
        local_conf dict - enables these parameters to be filtered from others
        which don't apply to this middleware
        @type prefix: str
        @param local_conf: attribute settings to apply
        @type local_conf: dict
        """
        cls = self.__class__
        self.resource_server_url = cls._get_config_option(prefix, local_conf,
                                                cls.RESOURCE_SERVER_URL_OPTION)
        self.session_env_key = cls._get_config_option(prefix, local_conf,
                                                      cls.SESSION_KEY_OPTION)
        self.token_env_key = self._get_config_option(prefix, local_conf,
                                                     cls.TOKEN_KEY_OPTION)
        self.certificate_env_key = self._get_config_option(prefix, local_conf,
                                            cls.CERTIFICATE_SESSION_KEY_OPTION)
        self.certificate_session_key = self._get_config_option(prefix,
                                local_conf, cls.CERTIFICATE_ENVIRON_KEY_OPTION)

        # SSL configuration
        client_cert = cls._get_config_option(prefix, local_conf, 'client_cert')
        client_key = cls._get_config_option(prefix, local_conf, 'client_key')
        ca_cert_file = cls._get_config_option(prefix, local_conf,
                                              'ca_cert_file')
        ca_dir = cls._get_config_option(prefix, local_conf, 'ca_dir')
        self.ssl_config = SSlContextConfig(client_key, client_cert,
                                           ca_cert_file, ca_dir, True)

        # OAuth client configuration
        self.certificate_request_parameter = cls._get_config_option(prefix,
                        local_conf, cls.CERTIFICATE_REQUEST_PARAMETER_OPTION)

    @classmethod
    def _get_config_option(cls, prefix, local_conf, key):
        value = local_conf.get(prefix + key, cls.PROPERTY_DEFAULTS.get(key,
                                                                      None))
        log.debug("CertificateRequestMiddleware configuration %s=%s", key,
                  value)
        return value

    @staticmethod
    def _get_http_status_string(status):
        return ("%d %s" % (status, httplib.responses[status]))

    @classmethod
    def filter_app_factory(cls, app, app_conf, **local_conf):
        return cls(app, app_conf, **local_conf)



"""OAuth 2.0 WSGI server middleware providing MyProxy certificates as access tokens
"""
__author__ = "R B Wilkinson"
__date__ = "28/02/12"
__copyright__ = "(C) 2012 Science and Technology Facilities Council"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id$"

import os

from genshi.template import TemplateLoader

from ndg.oauth.client.lib.render.renderer_interface import RendererInterface

class GenshiRenderer(RendererInterface):
    """Implementation of the renderer interface using Genshi
    """
    def render(self, filename, parameters):
        """Render a page from a template.
        @type filename: basestring
        @param filename: filename of template
        @type parameters: dict
        @param parameters: parameters to substitute into template
        @rtype: basestring
        @return: rendered template
        """
        fname = os.path.basename(filename)
        dirname = os.path.dirname(filename)
        loader = TemplateLoader(dirname, auto_reload=True)
        tmpl = loader.load(fname)
        response = tmpl.generate(c=parameters).render('html')
        return response



class RenderingConfiguration(object):
    """Holds configuration parameters for renderer.
    """
    __slots__ = ('__params')

    def __init__(self, pamameterNames, prefix, kw):
        """
        @type parameterNames: iterable
        @param parameterNames: names for which parameters are to be initialised
        @type prefix: basestring
        @param prefix: prefix matching start of relevant 
        @type kw: dict
        @param kw: key/value pairs of parameter names including the prefix and
        corresponding values
        """
        self.__params = dict.fromkeys(pamameterNames, '')
        prefixLength = len(prefix)
        for k, v in kw.iteritems():
            if k.startswith(prefix):
                self.__params[k[prefixLength:]] = v

    @property
    def parameters(self):
        return self.__params

    def merged_parameters(self, parameters):
        """Returns a copy of the parameters merged with specified additional
        ones.
        @type parameters: dict
        @param parameters: parameter name/value pairs to merge
        @rtype: dict
        @return: merged parameters
        """
        result = self.__params.copy()
        for k, v in parameters.iteritems():
            result[k] = v
        return result



__author__ = "R B Wilkinson"
__date__ = "29/02/12"
__copyright__ = "(C) 2012 Science and Technology Facilities Council"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id$"

try:
    from setuptools import setup, find_packages
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup, find_packages

_long_description = """\
This is an OAuth 2.0 client library and WSGI middleware filter.

It supports simple string-based bearer token and a custom extension to enable 
the use of X.509 certificates as tokens.  The latter has been added to enable
a SLCS (Short-lived Credential Service) to issue delegated X.509-based 
credentials using OAuth.

Releases
========
0.5.1
-----
New pseudo release to keep in sync with ndg_oauth_server package versioning. No
changes from 0.4.0.  New ndg_oauth_server 0.5.1 contains enhancements from W
van Engen including support for password based authentication for clients. See
ndg_oauth_server package for details.

0.4.0
-----
 * Revised examples in ndg.oauth.client.examples.  bearer_tok uses bearer token
   to secure access to a simple html page on a resource server, slcs is an 
   example protecting a short-lived credential service aka Online Certificate 
   Authority.  This requires the ContrailOnlineCAClient package and should be 
   used in conjunction with the equivalent example in the ndg_oauth_server 
   example.
 * Revised ndg.oauth.client.lib.oauth2client.Oauth2Client to include support for
   bearer access token passed in Authorization header to resource server.
 
Prerequisites
=============
This has been developed and tested for Python 2.6 and 2.7.

Installation
============
Installation can be performed using easy_install or pip.  

Configuration
=============
Examples are contained in the examples/ sub-folder:

bearer_tok/:
  This configures a simple test application that uses string based tokens.
slcs/:
  Bearer token example protecting a Short-Lived Credential Service or OnlineCA.
  ContrailOnlineCAService package is needed for this example.
  
The examples should be used in conjunction with the ndg_oauth_server package.
"""

setup(
    name =                      'ndg_oauth_client',
    version =                   '0.5.1',
    description =               'OAuth 2.0 client',
    long_description =          _long_description,
    author =                    'R. B. Wilkinson',
    maintainer =         	    'Philip Kershaw',
    maintainer_email =          'Philip.Kershaw@stfc.ac.uk',
    #url ='',
    license =                   'BSD - See LICENCE file for details',
    install_requires =[
        "PasteScript",
        "Beaker",
        "WebOb",
        "pyOpenSSL",
        "ndg_httpsclient",
        "pyasn1",
    ],
    packages =find_packages(),
    package_data =      {
        'ndg.oauth.client.templates': [
            '*.html', 'public/layout/*.css', 'public/layout/*.png',
            'public/layout/icons/*.png'
        ],
        'ndg.oauth.client.examples': [
            'README'
        ],
        'ndg.oauth.client.examples.bearer_tok': [
            'README', '*.ini'
        ],
        'ndg.oauth.client.examples.slcs': [
            'README', '*.ini'
        ],
        'ndg.oauth.client.examples.shared_config': [
            'README', 'pki/*.crt', 'pki/*.key', 'pki/ca/*.0'
        ]
    },
    zip_safe =False,
    classifiers = [
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Environment :: Web Environment',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Topic :: Security',
        'Topic :: Internet',
        'Topic :: Scientific/Engineering',
        'Topic :: System :: Distributed Computing',
        'Topic :: System :: Systems Administration :: Authentication/Directory',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
)



"""OAuth 2.0 WSGI server middleware providing MyProxy certificates as access tokens
"""
__author__ = "R B Wilkinson"
__date__ = "09/12/11"
__copyright__ = "(C) 2011 Science and Technology Facilities Council"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id$"

from OpenSSL import crypto

PRIKEY_NBITS = 4096
MESSAGE_DIGEST_TYPE = "md5"

def create_keypair(nBitsForKey=PRIKEY_NBITS):
    """Generate key pair and return as PEM encoded string
    @type nBitsForKey: int
    @param nBitsForKey: number of bits for private key generation - 
    default is 2048
    @rtype: OpenSSL.crypto.PKey
    @return: public/private key pair
    """
    keyPair = crypto.PKey()
    keyPair.generate_key(crypto.TYPE_RSA, nBitsForKey)
    
    return keyPair
        
def create_certreq(CN, keyPair, messageDigest=MESSAGE_DIGEST_TYPE):
    """Create a certificate request.
    
    @type CN: basestring
    @param CN: Common Name for certificate - effectively the same as the
    username for the MyProxy credential
    @type keyPair: string/None
    @param keyPair: public/private key pair
    @type messageDigest: basestring
    @param messageDigest: message digest type - default is MD5
    @rtype: base string
    @return certificate request PEM text and private key PEM text
    """
    
    # Check all required certificate request DN parameters are set                
    # Create certificate request
    certReq = crypto.X509Req()
    
    # Create public key object
    certReq.set_pubkey(keyPair)
    
    # Add the public key to the request
    certReq.sign(keyPair, messageDigest)
    
    derCertReq = crypto.dump_certificate_request(crypto.FILETYPE_ASN1, 
                                                 certReq)

    return derCertReq

def getKeyPairPrivateKey(keyPair):
    """Extracts the private key from a key pair.

    @type keyPair: string
    @param keyPair: public/private key pair
    @rtype: base string
    @return private key PEM text
    """
    return crypto.dump_privatekey(crypto.FILETYPE_PEM, keyPair)



"""NDG OAuth Client ndg.oauth namespace package for server code

This is a setuptools namespace_package.  DO NOT place any other
code in this file!  There is no guarantee that it will be installed
with easy_install.  See:

http://peak.telecommunity.com/DevCenter/setuptools#namespace-packages

... for details.
"""
__author__ = "P J Kershaw"
__date__ = "09/07/12"
__copyright__ = "(C) 2012 Science and Technology Facilities Council"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = '$Id$'

__import__('pkg_resources').declare_namespace(__name__)
